from unittest.mock import patch

import huawei_solar.register_names as rn
import huawei_solar.register_values as rv
import pytest
from huawei_solar.exceptions import DecodeError
from huawei_solar.register_values import GridCode
from pymodbus.register_read_message import ReadHoldingRegistersResponse


@pytest.mark.asyncio()
async def test_get_model_name(huawei_solar):
    result = await huawei_solar.get(rn.MODEL_NAME)
    assert result.value == "SUN2000-3KTL-L1"
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_invalid_model_name(huawei_solar):
    with (
        patch.object(
            huawei_solar,
            "_read_registers",
            return_value=ReadHoldingRegistersResponse(
                [
                    21333,
                    20018,
                    12336,
                    12333,
                    13131,
                    21580,
                    11596,
                    12544,
                    0,
                    0,
                    0,
                    0,
                    0,
                    226,
                    10370,
                ],
            ),
        ),
        pytest.raises(DecodeError),
    ):
        await huawei_solar.get("model_name")
        # invalid utf-8 sequence from here:
        # https://stackoverflow.com/questions/1301402/example-invalid-utf8-string


@pytest.mark.asyncio()
async def test_get_serial_number(huawei_solar):
    result = await huawei_solar.get(rn.SERIAL_NUMBER)
    assert result.value == "HV3021621085"
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_multiple(huawei_solar):
    result = await huawei_solar.get_multiple([rn.MODEL_NAME, rn.SERIAL_NUMBER])
    assert result[0].value == "SUN2000-3KTL-L1"
    assert result[0].unit is None

    assert result[1].value == "HV3021621085"
    assert result[1].unit is None


@pytest.mark.asyncio()
async def test_get_model_id(huawei_solar):
    result = await huawei_solar.get("model_id")
    assert result.value == 348
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_nb_pv_strings(huawei_solar):
    result = await huawei_solar.get("nb_pv_strings")
    assert result.value == 2
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_nb_mpp_tracks(huawei_solar):
    result = await huawei_solar.get("nb_mpp_tracks")
    assert result.value == 2
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_rated_power(huawei_solar):
    result = await huawei_solar.get("rated_power")
    assert result.value == 3000
    assert result.unit == "W"


@pytest.mark.asyncio()
async def test_get_p_max(huawei_solar):
    result = await huawei_solar.get("P_max")
    assert result.value == 3300
    assert result.unit == "W"


@pytest.mark.asyncio()
async def test_get_s_max(huawei_solar):
    result = await huawei_solar.get("S_max")
    assert result.value == 3300
    assert result.unit == "VA"


@pytest.mark.asyncio()
async def test_get_q_max_out(huawei_solar):
    result = await huawei_solar.get("Q_max_out")
    assert result.value == 1980
    assert result.unit == "var"


@pytest.mark.asyncio()
async def test_get_q_max_in(huawei_solar):
    result = await huawei_solar.get("Q_max_in")
    assert result.value == -1980
    assert result.unit == "var"


@pytest.mark.asyncio()
async def test_get_state_1(huawei_solar):
    result = await huawei_solar.get(rn.STATE_1)
    assert result.value == ["Standby"]
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_state_1_extra_bits_set(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0b0111_1100_0000_0000]),
    ):
        result = await huawei_solar.get(rn.STATE_1)
        assert result.value == []
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_state_2(huawei_solar):
    result = await huawei_solar.get(rn.STATE_2)
    assert result.value == ["Locked", "PV disconnected", "No DSP data collection"]
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_state_2_extra_bits_set(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0b0111_1111_1111_1000]),
    ):
        result = await huawei_solar.get(rn.STATE_2)

        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_state_3(huawei_solar):
    result = await huawei_solar.get(rn.STATE_3)
    assert result.value == ["On-grid", "Off-grid switch disabled"]
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_state_3_extra_bits_set(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse(
            [0b0111_1111_1111_1111, 0b0111_1111_1111_1111],
        ),
    ):
        result = await huawei_solar.get(rn.STATE_3)
        assert result.value, ["Off-grid", "Off-grid switch enabled"]
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_1_some(huawei_solar):
    result = await huawei_solar.get(rn.ALARM_1)
    expected_result = [
        rv.ALARM_CODES_1[1],
        rv.ALARM_CODES_1[256],
    ]
    assert result.value == expected_result
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_1_none(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0]),
    ):
        result = await huawei_solar.get(rn.ALARM_1)
        assert result.value == []
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_1_all(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0b1111_1111_1111_1111]),
    ):
        result = await huawei_solar.get(rn.ALARM_1)
        expected_result = list(rv.ALARM_CODES_1.values())
        assert result.value == expected_result
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_2_some(huawei_solar):
    result = await huawei_solar.get(rn.ALARM_2)
    expected_result = [
        rv.ALARM_CODES_2[2],
        rv.ALARM_CODES_2[512],
    ]
    assert result.value == expected_result
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_2_none(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0]),
    ):
        result = await huawei_solar.get(rn.ALARM_2)
        expected_result = []
        assert result.value == expected_result
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_2_all(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0b1111_1111_1111_1111]),
    ):
        result = await huawei_solar.get(rn.ALARM_2)
        expected_result = list(rv.ALARM_CODES_2.values())
        assert result.value == expected_result
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_3_some(huawei_solar):
    result = await huawei_solar.get(rn.ALARM_3)
    expected_result = list(rv.ALARM_CODES_3.values())[0:2] + list(rv.ALARM_CODES_3.values())[3:5]
    assert result.value == expected_result
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_3_almost_all(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0b0111_1111_1111_1111]),
    ):
        result = await huawei_solar.get(rn.ALARM_3)
        expected_result = list(rv.ALARM_CODES_3.values())[:-1]
        assert result.value == expected_result
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_alarm_3_3rd_octet_bits_set(huawei_solar):
    with patch.object(
        huawei_solar,
        "_read_registers",
        return_value=ReadHoldingRegistersResponse([0b0000_1110_0000_0000]),
    ):
        result = await huawei_solar.get(rn.ALARM_3)
        expected_result = list(rv.ALARM_CODES_3.values())[9:12]
        assert result.value == expected_result
        assert result.unit is None


@pytest.mark.asyncio()
async def test_get_pv_01_voltage(huawei_solar):
    result = await huawei_solar.get("pv_01_voltage")
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_pv_01_current(huawei_solar):
    result = await huawei_solar.get("pv_01_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_pv_02_voltage(huawei_solar):
    result = await huawei_solar.get("pv_02_voltage")
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_pv_02_current(huawei_solar):
    result = await huawei_solar.get("pv_02_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_pv_03_voltage(huawei_solar):
    result = await huawei_solar.get("pv_03_voltage")
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_pv_03_current(huawei_solar):
    result = await huawei_solar.get("pv_03_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_pv_04_voltage(huawei_solar):
    result = await huawei_solar.get("pv_04_voltage")
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_pv_04_current(huawei_solar):
    result = await huawei_solar.get("pv_04_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_(huawei_solar):
    result = await huawei_solar.get(rn.PV_01_VOLTAGE)
    assert result.value == 0.0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_input_power(huawei_solar):
    result = await huawei_solar.get(rn.INPUT_POWER)
    assert result.value == 0
    assert result.unit == "W"


@pytest.mark.asyncio()
async def test_get_grid_voltage(huawei_solar):
    result = await huawei_solar.get(rn.GRID_VOLTAGE)
    assert result.value == 0.0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_line_voltage_a_b(huawei_solar):
    result = await huawei_solar.get(rn.LINE_VOLTAGE_A_B)
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_line_voltage_b_c(huawei_solar):
    result = await huawei_solar.get(rn.LINE_VOLTAGE_B_C)
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_line_voltage_c_a(huawei_solar):
    result = await huawei_solar.get(rn.LINE_VOLTAGE_C_A)
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_line_phase_a_voltage(huawei_solar):
    result = await huawei_solar.get(rn.PHASE_A_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_line_phase_b_voltage(huawei_solar):
    result = await huawei_solar.get(rn.PHASE_B_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_line_phase_c_voltage(huawei_solar):
    result = await huawei_solar.get(rn.PHASE_C_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


@pytest.mark.asyncio()
async def test_get_grid_current(huawei_solar):
    result = await huawei_solar.get("grid_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_phase_a_current(huawei_solar):
    result = await huawei_solar.get("phase_A_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_phase_b_current(huawei_solar):
    result = await huawei_solar.get("phase_B_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_phase_c_current(huawei_solar):
    result = await huawei_solar.get("phase_C_current")
    assert result.value == 0
    assert result.unit == "A"


@pytest.mark.asyncio()
async def test_get_day_active_power_peak(huawei_solar):
    result = await huawei_solar.get("day_active_power_peak")
    assert result.value == 225
    assert result.unit == "W"


@pytest.mark.asyncio()
async def test_get_active_power(huawei_solar):
    result = await huawei_solar.get("active_power")
    assert result.value == 0
    assert result.unit == "W"


@pytest.mark.asyncio()
async def test_get_reactive_power(huawei_solar):
    result = await huawei_solar.get("reactive_power")
    assert result.value == 0
    assert result.unit == "var"


@pytest.mark.asyncio()
async def test_get_power_factor(huawei_solar):
    result = await huawei_solar.get("power_factor")
    assert result.value == 0.0
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_grid_frequency(huawei_solar):
    result = await huawei_solar.get("grid_frequency")
    assert result.value == 0.0
    assert result.unit == "Hz"


@pytest.mark.asyncio()
async def test_get_efficiency(huawei_solar):
    result = await huawei_solar.get("efficiency")
    assert result.value == 0.0
    assert result.unit == "%"


@pytest.mark.asyncio()
async def test_get_internal_temperature(huawei_solar):
    result = await huawei_solar.get("internal_temperature")
    assert result.value == 0.0
    assert result.unit == "°C"


@pytest.mark.asyncio()
async def test_get_insulation_resistance(huawei_solar):
    result = await huawei_solar.get(rn.INSULATION_RESISTANCE)
    assert result.value == 3.0
    assert result.unit == "MOhm"


@pytest.mark.asyncio()
async def test_get_device_status(huawei_solar):
    result = await huawei_solar.get(rn.DEVICE_STATUS)
    assert result.value == "Standby: no irradiation"
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_device_status_invalid(huawei_solar):
    with (
        patch.object(
            huawei_solar,
            "_read_registers",
            return_value=ReadHoldingRegistersResponse([0b0000_0010_1111_1111]),
        ),
        pytest.raises(DecodeError),
    ):
        await huawei_solar.get(rn.DEVICE_STATUS)


@pytest.mark.asyncio()
async def test_get_fault_code(huawei_solar):
    result = await huawei_solar.get(rn.FAULT_CODE)
    assert result.value == 0
    assert result.unit is None


# @pytest.mark.asyncio()
# async def test_get_startup_time(huawei_solar):
#     result = await huawei_solar.get(rn.STARTUP_TIME)

#     assert result.value == datetime(2022, 1, 23, 10, 3, 49)
#     assert result.unit is None


# @pytest.mark.asyncio()
# async def test_get_shutdown_time(huawei_solar):
#     result = await huawei_solar.get(rn.SHUTDOWN_TIME)
#     assert result.value == datetime(2022, 1, 23, 18, 7, 25)
#     assert result.unit is None


@pytest.mark.asyncio()
async def test_get_accumulated_yield_energy(huawei_solar):
    result = await huawei_solar.get(rn.ACCUMULATED_YIELD_ENERGY)
    assert result.value == 207.34
    assert result.unit == "kWh"


@pytest.mark.asyncio()
async def test_get_daily_yield_energy(huawei_solar):
    result = await huawei_solar.get(rn.DAILY_YIELD_ENERGY)
    assert result.value == 0.65
    assert result.unit == "kWh"


@pytest.mark.asyncio()
async def test_get_nb_optimizers(huawei_solar):
    result = await huawei_solar.get(rn.NB_OPTIMIZERS)
    assert result.value == 10
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_nb_online_optimizers(huawei_solar):
    result = await huawei_solar.get(rn.NB_ONLINE_OPTIMIZERS)
    assert result.value == 0
    assert result.unit is None


# @pytest.mark.asyncio()
# async def test_get_system_time(huawei_solar):
#     result = await huawei_solar.get(rn.SYSTEM_TIME)
#     tmp = datetime(2022, 1, 23, 23, 6, 35)
#     assert result.value == tmp
#     assert result.unit is None


@pytest.mark.asyncio()
async def test_get_grid_code(huawei_solar):
    result = await huawei_solar.get(rn.GRID_CODE)
    expected_result = GridCode(standard="C10/11", country="Belgium")
    assert result.value == expected_result
    assert result.unit is None


@pytest.mark.asyncio()
async def test_get_time_zone(huawei_solar):
    result = await huawei_solar.get(rn.TIME_ZONE)
    assert result.value == 60
    assert result.unit == "min"



"""Generic Utilities."""

from datetime import datetime, tzinfo


def get_local_timezone() -> tzinfo:
    """Return the current local timezone."""
    local_timezone = datetime.now().astimezone().tzinfo

    if not local_timezone:
        raise ValueError
    return local_timezone



"""check if the register config is correct"""

import logging

from huawei_solar.registers import REGISTERS, TargetDevice

_LOGGER = logging.getLogger(__name__)


def test_register_config():
    """Parse all REGISTERS and check for correct order and potential overlaps"""
    registers = [r for r in REGISTERS.values() if TargetDevice.SUN2000 in r.target_device]
    registers.sort(key=lambda x: x.register)

    for idx in range(1, len(registers)):
        if registers[idx].register in [32066, 32072, 40000, 47028, 47255]:
            # skip these registers, as they have multiple entries
            continue
        if registers[idx - 1].register + registers[idx - 1].length > registers[idx].register:
            raise ValueError(
                f"Requested registers must be in monotonically increasing order, "
                f"but {registers[idx-1].register} + {registers[idx-1].length} > {registers[idx].register}!",
            )
        if registers[idx - 1].register + registers[idx - 1].length < registers[idx].register:
            _LOGGER.info(
                "There is a gap between %s and %s!",
                {registers[idx - 1].register},
                {registers[idx].register},
            )


def test_register_config_emma():
    """Parse all REGISTERS and check for correct order and potential overlaps"""
    registers = [r for r in REGISTERS.values() if TargetDevice.EMMA in r.target_device]
    registers.sort(key=lambda x: x.register)

    for idx in range(1, len(registers)):
        if registers[idx].register in []:
            # skip these registers, as they have multiple entries
            continue
        if registers[idx - 1].register + registers[idx - 1].length > registers[idx].register:
            raise ValueError(
                f"Requested registers must be in monotonically increasing order, "
                f"but {registers[idx-1].register} + {registers[idx-1].length} > {registers[idx].register}!",
            )
        if registers[idx - 1].register + registers[idx - 1].length < registers[idx].register:
            _LOGGER.info(
                "There is a gap between %s and %s!",
                {registers[idx - 1].register},
                {registers[idx].register},
            )



"""Definitions of register values returned by the Huawei inverter."""

from enum import IntEnum
from typing import NamedTuple


class GridCode(NamedTuple):
    """GridCode."""

    standard: str
    country: str


class Alarm(NamedTuple):
    """Alarm."""

    name: str
    id: int
    level: str


class OnOffBit(NamedTuple):
    """Bit with different meaning when on or off."""

    off_value: str
    on_value: str


DEVICE_STATUS_DEFINITIONS = {
    0x0000: "Standby: initializing",
    0x0001: "Standby: detecting insulation resistance",
    0x0002: "Standby: detecting irradiation",
    0x0003: "Standby: grid detecting",
    0x0100: "Starting",
    0x0200: "On-grid",
    0x0201: "Grid Connection: power limited",
    0x0202: "Grid Connection: self-derating",
    0x0203: "Off-grid mode: running",
    0x0300: "Shutdown: fault",
    0x0301: "Shutdown: command",
    0x0302: "Shutdown: OVGR",
    0x0303: "Shutdown: communication disconnected",
    0x0304: "Shutdown: power limited",
    0x0305: "Shutdown: manual startup required",
    0x0306: "Shutdown: DC switches disconnected",
    0x0307: "Shutdown: rapid cutoff",
    0x0308: "Shutdown: input underpowered",
    0x0401: "Grid scheduling: cosphi-P curve",
    0x0402: "Grid scheduling: Q-U curve",
    0x0403: "Grid scheduling: PF-U curve",
    0x0404: "Grid scheduling: dry contact",
    0x0405: "Grid scheduling: Q-P curve",
    0x0500: "Spot-check ready",
    0x0501: "Spot-checking",
    0x0600: "Inspecting",
    0x0700: "AFCI self check",
    0x0800: "I-V scanning",
    0x0900: "DC input detection",
    0x0A00: "Running: off-grid charging",
    0xA000: "Standby: no irradiation",
}


class _IntEnumWithPrettyString(IntEnum):
    def __str__(self) -> str:
        """Return pretty string representation."""
        return self.name.replace("_", " ").capitalize()


class StorageStatus(_IntEnumWithPrettyString):
    """Status of the attached energy storage."""

    OFFLINE = 0
    STANDBY = 1
    RUNNING = 2
    FAULT = 3
    SLEEP_MODE = 4


class StorageWorkingModesA(_IntEnumWithPrettyString):
    """Working mode of the Connected Energy Storage."""

    UNLIMITED = 0
    GRID_CONNECTION_WITH_ZERO_POWER = 1
    GRID_CONNECTION_WITH_LIMITED_POWER = 2


class StorageWorkingModesB(_IntEnumWithPrettyString):
    """Working mode of the Connected Energy Storage."""

    NONE = 0
    FORCIBLE_CHARGE_DISCHARGE = 1
    TIME_OF_USE_LG = 2
    FIXED_CHARGE_DISCHARGE = 3
    MAXIMISE_SELF_CONSUMPTION = 4
    FULLY_FED_TO_GRID = 5
    TIME_OF_USE_LUNA2000 = 6
    REMOTE_SCHEDULING_MAXIMUM_SELF_USE = 7
    REMOTE_SCHEDULING_FULL_INTERNET_ACCESS = 8
    REMOTE_SCHEDULING_TOU = 9
    AI_ENERGY_MANAGEMENT_AND_SCHEDULING = 10
    REMOTE_SCHEDULING_AI_CONTROL = 11


class StorageWorkingModesC(_IntEnumWithPrettyString):
    """Working mode of the Connected Energy Storage."""

    ADAPTIVE = 0
    FIXED_CHARGE_DISCHARGE = 1
    MAXIMISE_SELF_CONSUMPTION = 2
    TIME_OF_USE_LG = 3
    FULLY_FED_TO_GRID = 4
    TIME_OF_USE_LUNA2000 = 5


class StorageProductModel(IntEnum):
    """Storage Product Model."""

    NONE = 0
    LG_RESU = 1
    HUAWEI_LUNA2000 = 2


class StorageForcibleChargeDischarge(_IntEnumWithPrettyString):
    """Storage Forcible charge/discharge mode."""

    STOP = 0
    CHARGE = 1
    DISCHARGE = 2


class StorageForcibleChargeDischargeTargetMode(_IntEnumWithPrettyString):
    """Storage Forcible charge/discharge target mode."""

    TIME = 0
    SOC = 1


class StorageExcessPvEnergyUseInTOU(_IntEnumWithPrettyString):
    """Storage Excess PV Energy use in Time-of-Use."""

    FED_TO_GRID = 0
    CHARGE = 1


class ActivePowerControlMode(_IntEnumWithPrettyString):
    """Active Power Control Mode."""

    UNLIMITED = 0  # default mode
    DI_ACTIVE_SCHEDULING = 1
    ZERO_POWER_GRID_CONNECTION = 5
    POWER_LIMITED_GRID_CONNECTION_WATT = 6
    POWER_LIMITED_GRID_CONNECTION_PERCENT = 7


class MeterStatus(_IntEnumWithPrettyString):
    """Power meter status."""

    OFFLINE = 0
    NORMAL = 1


class MeterType(_IntEnumWithPrettyString):
    """Power meter type."""

    SINGLE_PHASE = 0
    THREE_PHASE = 1


class MeterTypeCheck(_IntEnumWithPrettyString):
    """Power meter type check."""

    RECOGNIZING = 0
    MATCHES_WITH_METER = 1
    MATCHES_NOT_WITH_METER = 2


class BackupVoltageIndependentOperation(IntEnum):
    """Backup voltage independent operation."""

    BV_101V = 0
    BV_202V = 1


GRID_CODES = {
    0: GridCode("VDE-AR-N-4105", "Germany"),
    1: GridCode("NB/T 32004", "China"),
    2: GridCode("UTE C 15-712-1(A)", "France"),
    3: GridCode("UTE C 15-712-1(B)", "France"),
    4: GridCode("UTE C 15-712-1(C)", "France"),
    5: GridCode("VDE 0126-1-1-BU", "Bulgary"),
    6: GridCode("VDE 0126-1-1-GR(A)", "Greece"),
    7: GridCode("VDE 0126-1-1-GR(B)", "Greece"),
    8: GridCode("BDEW-MV", "Germany"),
    9: GridCode("G59-England", "UK"),
    10: GridCode("G59-Scotland", "UK"),
    11: GridCode("G83-England", "UK"),
    12: GridCode("G83-Scotland", "UK"),
    13: GridCode("CEI0-21", "Italy"),
    14: GridCode("EN50438-CZ", "Czech Republic"),
    15: GridCode("RD1699/661", "Spain"),
    16: GridCode("RD1699/661-MV480", "Spain"),
    17: GridCode("EN50438-NL", "Netherlands"),
    18: GridCode("C10/11", "Belgium"),
    19: GridCode("AS4777", "Australia"),
    20: GridCode("IEC61727", "General"),
    21: GridCode("Custom (50 Hz)", "Custom"),
    22: GridCode("Custom (60 Hz)", "Custom"),
    23: GridCode("CEI0-16", "Italy"),
    24: GridCode("CHINA-MV480", "China"),
    25: GridCode("CHINA-MV", "China"),
    26: GridCode("TAI-PEA", "Thailand"),
    27: GridCode("TAI-MEA", "Thailand"),
    28: GridCode("BDEW-MV480", "Germany"),
    29: GridCode("Custom MV480 (50 Hz)", "Custom"),
    30: GridCode("Custom MV480 (60 Hz)", "Custom"),
    31: GridCode("G59-England-MV480", "UK"),
    32: GridCode("IEC61727-MV480", "General"),
    33: GridCode("UTE C 15-712-1-MV480", "France"),
    34: GridCode("TAI-PEA-MV480", "Thailand"),
    35: GridCode("TAI-MEA-MV480", "Thailand"),
    36: GridCode("EN50438-DK-MV480", "Denmark"),
    37: GridCode("Japan standard (50 Hz)", "Japan"),
    38: GridCode("Japan standard (60 Hz)", "Japan"),
    39: GridCode("EN50438-TR-MV480", "Turkey"),
    40: GridCode("EN50438-TR", "Turkey"),
    41: GridCode("C11/C10-MV480", "Belgium"),
    42: GridCode("Philippines", "Philippines"),
    43: GridCode("Philippines-MV480", "Philippines"),
    44: GridCode("AS4777-MV480", "Australia"),
    45: GridCode("NRS-097-2-1", "South Africa"),
    46: GridCode("NRS-097-2-1-MV480", "South Africa"),
    47: GridCode("KOREA", "South Korea"),
    48: GridCode("IEEE 1547-MV480", "USA"),
    49: GridCode("IEC61727-60Hz", "General"),
    50: GridCode("IEC61727-60Hz-MV480", "General"),
    51: GridCode("CHINA_MV500", "China"),
    52: GridCode("ANRE", "Romania"),
    53: GridCode("ANRE-MV480", "Romania"),
    54: GridCode("ELECTRIC RULE NO.21-MV480", "California, USA"),
    55: GridCode("HECO-MV480", "Hawaii, USA"),
    56: GridCode("PRC_024_Eastern-MV480", "Eastern USA"),
    57: GridCode("PRC_024_Western-MV480", "Western USA"),
    58: GridCode("PRC_024_Quebec-MV480", "Quebec, Canada"),
    59: GridCode("PRC_024_ERCOT-MV480", "Texas, USA"),
    60: GridCode("PO12.3-MV480", "Spain"),
    61: GridCode("EN50438_IE-MV480", "Ireland"),
    62: GridCode("EN50438_IE", "Ireland"),
    63: GridCode("IEEE 1547a-MV480", "USA"),
    64: GridCode("Japan standard (MV420-50 Hz)", "Japan"),
    65: GridCode("Japan standard (MV420-60 Hz)", "Japan"),
    66: GridCode("Japan standard (MV440-50 Hz)", "Japan"),
    67: GridCode("Japan standard (MV440-60 Hz)", "Japan"),
    68: GridCode("IEC61727-50Hz-MV500", "General"),
    70: GridCode("CEI0-16-MV480", "Italy"),
    71: GridCode("PO12.3", "Spain"),
    72: GridCode("Japan standard (MV400-50 Hz)", "Japan"),
    73: GridCode("Japan standard (MV400-60 Hz)", "Japan"),
    74: GridCode("CEI0-21-MV480", "Italy"),
    75: GridCode("KOREA-MV480", "South Korea"),
    76: GridCode("Egypt ETEC", "Egypt"),
    77: GridCode("Egypt ETEC-MV480", "Egypt"),
    78: GridCode("CHINA_MV800", "China"),
    79: GridCode("IEEE 1547-MV600", "USA"),
    80: GridCode("ELECTRIC RULE NO.21-MV600", "California, USA"),
    81: GridCode("HECO-MV600", "Hawaii, USA"),
    82: GridCode("PRC_024_Eastern-MV600", "Eastern USA"),
    83: GridCode("PRC_024_Western-MV600", "Western USA"),
    84: GridCode("PRC_024_Quebec-MV600", "Quebec, Canada"),
    85: GridCode("PRC_024_ERCOT-MV600", "Texas, USA"),
    86: GridCode("IEEE 1547a-MV600", "USA"),
    87: GridCode("EN50549-LV", "Ireland"),
    88: GridCode("EN50549-MV480", "Ireland"),
    89: GridCode("Jordan-Transmission", "Jordan"),
    90: GridCode("Jordan-Transmission-MV480", "Jordan"),
    91: GridCode("NAMIBIA", "Namibia"),
    92: GridCode("ABNT NBR 16149", "Brazil"),
    93: GridCode("ABNT NBR 16149-MV480", "Brazil"),
    94: GridCode("SA_RPPs", "South Africa"),
    95: GridCode("SA_RPPs-MV480", "South Africa"),
    96: GridCode("INDIA", "India"),
    97: GridCode("INDIA-MV500", "India"),
    98: GridCode("ZAMBIA", "Zambia"),
    99: GridCode("ZAMBIA-MV480", "Zambia"),
    100: GridCode("Chile", "Chile"),
    101: GridCode("Chile-MV480", "Chile"),
    102: GridCode("CHINA-MV500-STD", "China"),
    103: GridCode("CHINA-MV480-STD", "China"),
    104: GridCode("Mexico-MV480", "Mexico"),
    105: GridCode("Malaysian", "Malaysia"),
    106: GridCode("Malaysian-MV480", "Malaysia"),
    107: GridCode("KENYA_ETHIOPIA", "East Africa"),
    108: GridCode("KENYA_ETHIOPIA-MV480", "East Africa"),
    109: GridCode("G59-England-MV800", "UK"),
    110: GridCode("NIGERIA", "Nigeria"),
    111: GridCode("NIGERIA-MV480", "Nigeria"),
    112: GridCode("DUBAI", "Dubai"),
    113: GridCode("DUBAI-MV480", "Dubai"),
    114: GridCode("Northern Ireland", "Northern Ireland"),
    115: GridCode("Northern Ireland-MV480", "Northern Ireland"),
    116: GridCode("Cameroon", "Cameroon"),
    117: GridCode("Cameroon-MV480", "Cameroon"),
    118: GridCode("Jordan Distribution", "Jordan"),
    119: GridCode("Jordan Distribution-MV480", "Jordan"),
    120: GridCode("Custom MV600-50Hz", "Custom"),
    121: GridCode("AS4777-MV800", "Australia"),
    122: GridCode("INDIA-MV800", "India"),
    123: GridCode("IEC61727-MV800", "General"),
    124: GridCode("BDEW-MV800", "Germany"),
    125: GridCode("ABNT NBR 16149-MV800", "Brazil"),
    126: GridCode("UTE C 15-712-1-MV800", "France"),
    127: GridCode("Chile-MV800", "Chile"),
    128: GridCode("Mexico-MV800", "Mexico"),
    129: GridCode("EN50438-TR-MV800", "Turkey"),
    130: GridCode("TAI-PEA-MV800", "Thailand"),
    131: GridCode("Philippines-MV800", "Philippines"),
    132: GridCode("Malaysian-MV800", "Malaysia"),
    133: GridCode("NRS-097-2-1-MV800", "South Africa"),
    134: GridCode("SA_RPPs-MV800", "South Africa"),
    135: GridCode("Jordan-Transmission-MV800", "Jordan"),
    136: GridCode("Jordan-Distribution-MV800", "Jordan"),
    137: GridCode("Egypt ETEC-MV800", "Egypt"),
    138: GridCode("DUBAI-MV800", "Dubai"),
    139: GridCode("SAUDI-MV800", "Saudi Arabia"),
    140: GridCode("EN50438_IE-MV800", "Ireland"),
    141: GridCode("EN50549-MV800", "Ireland"),
    142: GridCode("Northern Ireland-MV800", "Northern Ireland"),
    143: GridCode("CEI0-21-MV800", "Italy"),
    144: GridCode("IEC 61727-MV800-60Hz", "General"),
    145: GridCode("NAMIBIA_MV480", "Namibia"),
    146: GridCode("Japan (LV202-50Hz)", "Japan"),
    147: GridCode("Japan (LV202-60Hz)", "Japan"),
    148: GridCode("Pakistan-MV800", "Pakistan"),
    149: GridCode("BRASIL-ANEEL-MV800", "Brazil"),
    150: GridCode("Israel-MV800", "Israel"),
    151: GridCode("CEI0-16-MV800", "Italy"),
    152: GridCode("ZAMBIA-MV800", "Zambia"),
    153: GridCode("KENYA_ETHIOPIA-MV800", "East Africa"),
    154: GridCode("NAMIBIA_MV800", "Namibia"),
    155: GridCode("Cameroon-MV800", "Cameroon"),
    156: GridCode("NIGERIA-MV800", "Nigeria"),
    157: GridCode("ABUDHABI-MV800", "Abu Dhabi"),
    158: GridCode("LEBANON", "Lebanon"),
    159: GridCode("LEBANON-MV480", "Lebanon"),
    160: GridCode("LEBANON-MV800", "Lebanon"),
    161: GridCode("ARGENTINA-MV800", "Argentina"),
    162: GridCode("ARGENTINA-MV500", "Argentina"),
    163: GridCode("Jordan-Transmission-HV", "Jordan"),
    164: GridCode("Jordan-Transmission-HV480", "Jordan"),
    165: GridCode("Jordan-Transmission-HV800", "Jordan"),
    166: GridCode("TUNISIA", "Tunisia"),
    167: GridCode("TUNISIA-MV480", "Tunisia"),
    168: GridCode("TUNISIA-MV800", "Tunisia"),
    169: GridCode("JAMAICA-MV800", "Jamaica"),
    170: GridCode("AUSTRALIA-NER", "Australia"),
    171: GridCode("AUSTRALIA-NER-MV480", "Australia"),
    172: GridCode("AUSTRALIA-NER-MV800", "Australia"),
    173: GridCode("SAUDI", "Saudi Arabia"),
    174: GridCode("SAUDI-MV480", "Saudi Arabia"),
    175: GridCode("Ghana-MV480", "Ghana"),
    176: GridCode("Israel", "Israel"),
    177: GridCode("Israel-MV480", "Israel"),
    178: GridCode("Chile-PMGD", "Chile"),
    179: GridCode("Chile-PMGD-MV480", "Chile"),
    180: GridCode("VDE-AR-N4120-HV", "Germany"),
    181: GridCode("VDE-AR-N4120-HV480", "Germany"),
    182: GridCode("VDE-AR-N4120-HV800", "Germany"),
    183: GridCode("IEEE 1547-MV800", "USA"),
    184: GridCode("Nicaragua-MV800", "Nicaragua"),
    185: GridCode("IEEE 1547a-MV800", "USA"),
    186: GridCode("ELECTRIC RULE NO.21-MV800", "California, USA"),
    187: GridCode("HECO-MV800", "Hawaii, USA"),
    188: GridCode("PRC_024_Eastern-MV800", "Eastern USA"),
    189: GridCode("PRC_024_Western-MV800", "Western USA"),
    190: GridCode("PRC_024_Quebec-MV800", "Quebec, Canada"),
    191: GridCode("PRC_024_ERCOT-MV800", "Texas, USA"),
    192: GridCode("Custom-MV800-50Hz", "Custom"),
    193: GridCode("RD1699/661-MV800", "Spain"),
    194: GridCode("PO12.3-MV800", "Spain"),
    195: GridCode("Mexico-MV600", "Mexico"),
    196: GridCode("Vietnam-MV800", "Vietnam"),
    197: GridCode("CHINA-LV220/380", "China"),
    198: GridCode("SVG-LV", "Dedicated"),
    199: GridCode("Vietnam", "Vietnam"),
    200: GridCode("Vietnam-MV480", "Vietnam"),
    201: GridCode("Chile-PMGD-MV800", "Chile"),
    202: GridCode("Ghana-MV800", "Ghana"),
    203: GridCode("TAIPOWER", "Taiwan"),
    204: GridCode("TAIPOWER-MV480", "Taiwan"),
    205: GridCode("TAIPOWER-MV800", "Taiwan"),
    206: GridCode("IEEE 1547-LV208", "USA"),
    207: GridCode("IEEE 1547-LV240", "USA"),
    208: GridCode("IEEE 1547a-LV208", "USA"),
    209: GridCode("IEEE 1547a-LV240", "USA"),
    210: GridCode("ELECTRIC RULE NO.21-LV208", "USA"),
    211: GridCode("ELECTRIC RULE NO.21-LV240", "USA"),
    212: GridCode("HECO-O+M+H-LV208", "USA"),
    213: GridCode("HECO-O+M+H-LV240", "USA"),
    214: GridCode("PRC_024_Eastern-LV208", "USA"),
    215: GridCode("PRC_024_Eastern-LV240", "USA"),
    216: GridCode("PRC_024_Western-LV208", "USA"),
    217: GridCode("PRC_024_Western-LV240", "USA"),
    218: GridCode("PRC_024_ERCOT-LV208", "USA"),
    219: GridCode("PRC_024_ERCOT-LV240", "USA"),
    220: GridCode("PRC_024_Quebec-LV208", "USA"),
    221: GridCode("PRC_024_Quebec-LV240", "USA"),
    222: GridCode("ARGENTINA-MV480", "Argentina"),
    223: GridCode("Oman", "Oman"),
    224: GridCode("Oman-MV480", "Oman"),
    225: GridCode("Oman-MV800", "Oman"),
    226: GridCode("Kuwait", "Kuwait"),
    227: GridCode("Kuwait-MV480", "Kuwait"),
    228: GridCode("Kuwait-MV800", "Kuwait"),
    229: GridCode("Bangladesh", "Bangladesh"),
    230: GridCode("Bangladesh-MV480", "Bangladesh"),
    231: GridCode("Bangladesh-MV800", "Bangladesh"),
    232: GridCode("Chile-Net_Billing", "Chile"),
    233: GridCode("EN50438-NL-MV480", "Netherlands"),
    234: GridCode("Bahrain", "Bahrain"),
    235: GridCode("Bahrain-MV480", "Bahrain"),
    236: GridCode("Bahrain-MV800", "Bahrain"),
    237: GridCode("Fuel-Engine-Grid", "Dedicated"),
    238: GridCode("Japan-MV550-50Hz", "Japan"),
    239: GridCode("Japan-MV550-60Hz", "Japan"),
    241: GridCode("ARGENTINA", "Argentina"),
    242: GridCode("KAZAKHSTAN-MV800", "Kazakhstan"),
    243: GridCode("Mauritius", "Mauritius"),
    244: GridCode("Mauritius-MV480", "Mauritius"),
    245: GridCode("Mauritius-MV800", "Mauritius"),
    246: GridCode("Oman-PDO-MV800", "Oman"),
    247: GridCode("EN50438-SE", "Sweden"),
    248: GridCode("TAI-MEA-MV800", "Thailand"),
    249: GridCode("Pakistan", "Pakistan"),
    250: GridCode("Pakistan-MV480", "Pakistan"),
    251: GridCode("PORTUGAL-MV800", "Portugal"),
    252: GridCode("HECO-L+M-LV208", "USA"),
    253: GridCode("HECO-L+M-LV240", "USA"),
    254: GridCode("C10/11-MV800", "Belgium"),
    255: GridCode("Austria", "Austria"),
    256: GridCode("Austria-MV480", "Austria"),
    257: GridCode("G98", "UK"),
    258: GridCode("G99-TYPEA-LV", "UK"),
    259: GridCode("G99-TYPEB-LV", "UK"),
    260: GridCode("G99-TYPEB-HV", "UK"),
    261: GridCode("G99-TYPEB-HV-MV480", "UK"),
    262: GridCode("G99-TYPEB-HV-MV800", "UK"),
    263: GridCode("G99-TYPEC-HV-MV800", "UK"),
    264: GridCode("G99-TYPED-MV800", "UK"),
    265: GridCode("G99-TYPEA-HV", "UK"),
    266: GridCode("CEA-MV800", "India"),
    267: GridCode("EN50549-MV400", "Europe"),
    268: GridCode("VDE-AR-N4110", "Germany"),
    269: GridCode("VDE-AR-N4110-MV480", "Germany"),
    270: GridCode("VDE-AR-N4110-MV800", "Germany"),
    271: GridCode("Panama-MV800", "Panama"),
    272: GridCode("Macedonia-MV800", "Macedonia"),
    273: GridCode("NTS", "Spain"),
    274: GridCode("NTS-MV480", "Spain"),
    275: GridCode("NTS-MV800", "Spain"),
    276: GridCode("AS4777-WP", "Australia"),
    277: GridCode("CEA", "India"),
    278: GridCode("CEA-MV480", "India"),
    279: GridCode("SINGAPORE", "Singapore"),
    280: GridCode("SINGAPORE-MV480", "Singapore"),
    281: GridCode("SINGAPORE-MV800", "Singapore"),
    282: GridCode("HONGKONG", "Hong Kong"),
    283: GridCode("HONGKONG-MV480", "Hong Kong"),
    284: GridCode("C10/11-MV400", "Belgium"),
    285: GridCode("KOREA-MV800", "Korea"),
    286: GridCode("Cambodia", "Cambodia"),
    287: GridCode("Cambodia-MV480", "Cambodia"),
    288: GridCode("Cambodia-MV800", "Cambodia"),
    289: GridCode("EN50549-SE", "Sweden"),
    290: GridCode("GREG030", "Columbia"),
    291: GridCode("GREG030-MV440", "Columbia"),
    292: GridCode("GREG030-MV480", "Columbia"),
    293: GridCode("GREG030-MV800", "Columbia"),
    294: GridCode("PERU-MV800", "Peru"),
    295: GridCode("PORTUGAL", "Portugal"),
    296: GridCode("PORTUGAL-MV480", "Portugal"),
    297: GridCode("AS4777-ACT", "Australia"),
    298: GridCode("AS4777-NSW-ESS", "Australia"),
    299: GridCode("AS4777-NSW-AG", "Australia"),
    300: GridCode("AS4777-QLD", "Australia"),
    301: GridCode("AS4777-SA", "Australia"),
    302: GridCode("AS4777-VIC", "Australia"),
    303: GridCode("EN50549-PL", "Polamd"),
    304: GridCode("Island-Grid", "General"),
    305: GridCode("TAIPOWER-LV220", "China Taiwan"),
    306: GridCode("Mexico-LV220", "Mexico"),
    307: GridCode("ABNT NBR 161490LV127", "Brazil"),
    308: GridCode("Philippines-LV220-50Hz", "Philippines"),
    309: GridCode("Philippines-LV220-60Hz", "Philippines"),
    310: GridCode("Israel-HV800", "Israel"),
    311: GridCode("DENMARK-EN50549-DK1-LV230", "Denmark"),
    312: GridCode("DENMARK-EN50549-DK2-LV230", "Denmark"),
    313: GridCode("SWITZERLAND-NA/EEA:2020-LV230", "Switzerland"),
    314: GridCode("Japan-LV202-50Hz", "Japan"),
    315: GridCode("Japan-LC202-60Hz", "Japan"),
    316: GridCode("AUSTRIA-MV800", "Austria"),
    317: GridCode("AUSTRIA-HV800", "Austria"),
    318: GridCode("POLAND-EN50549-MV800", "Poland"),
    319: GridCode("IRELAND-EN50549-LV230", "Ireland"),
    320: GridCode("IRELAND-EN50549-MV480", "Ireland"),
    321: GridCode("IRELAND-EN50549-MV800", "Ireland"),
    322: GridCode("DENMARK-EN50549-MV800", "Denmark"),
    323: GridCode("FRANCE-RTE-MV800", "France"),
    324: GridCode("AUSTRALIA-AS4777_A-LV230", "Australia"),
    325: GridCode("AUSTRALIA-AS4777_B-LV230", "Australia"),
    326: GridCode("AUSTRALIA-AS4777_C-LV230", "Australia"),
    327: GridCode("AUSTRALIA-AS4777_NZ-LV230", "Australia"),
    328: GridCode("AUSTRALIA-AS4777_A-MV800", "Australia"),
    329: GridCode("CHINA-GBT34120-MV800", "China"),
    330: GridCode("UZBEKISTAN-MV800", "Uzbekistan"),
    331: GridCode("CHINA-GBT34120-MV380", "China"),
    333: GridCode("CHINA-MV690", "China"),
    334: GridCode("IEC61727-MV720", "India"),
    335: GridCode("INDIA-CEA-MV720", "India"),
    336: GridCode("SA-NRS-097-MV720", "South Africa"),
    337: GridCode("SA-RPPS-MV720", "South Africa"),
    338: GridCode("SAUDI-MV720", "Saudi Arabia"),
    339: GridCode("UZBEKISTAN-MV720", "Uzbekistan"),
    340: GridCode("EGYPT-ETEC-MC720", "Egypt"),
    341: GridCode("KAZAKHSTAN-MV690", "Kazakhstan"),
    342: GridCode("CHINA-CUSTOM-MV800", "China"),
    343: GridCode("JAPAN-MV200-50Hz", "Japan"),
    344: GridCode("JAPAN-MV210-50Hz", "Japan"),
    345: GridCode("JAPAN-MV230-50Hz", "Japan"),
    346: GridCode("JAPAN-MV250-50Hz", "Japan"),
    347: GridCode("JAPAN-MV200-60Hz", "Japan"),
    348: GridCode("JAPAN-MV210-60Hz", "Japan"),
    349: GridCode("JAPAN-MV230-60Hz", "Japan"),
    350: GridCode("JAPAN-MV250-60Hz", "Japan"),
    351: GridCode("CZECH-EN50549-LV230", "Czech Republic"),
    352: GridCode("CZECH-EN50549-MV480", "Czech Republic"),
    353: GridCode("CZECH-EN50549-MV800", "Czech Republic"),
    354: GridCode("CHINA_MV315", "China"),
    355: GridCode("KOREA-MV690-60Hz", "Korea"),
    356: GridCode("ANRE-MV800", "Romania"),
    357: GridCode("FINLAND-EN50549-LV230", "Finland"),
    358: GridCode("AUSTRIA-MV400-50Hz", "Austria"),
    359: GridCode("SAUDI-LV220", "Saudi Arabia"),
    360: GridCode("CHINA-MV285", "China"),
    361: GridCode("CHINA-MV360", "China"),
    362: GridCode("JAPAN-MV270-50Hz", "Japan"),
    363: GridCode("JAPAN-MV330-50Hz", "Japan"),
    364: GridCode("JAPAN-MV380-50Hz", "Japan"),
    365: GridCode("JAPAN-MV270-60Hz", "Japan"),
    366: GridCode("JAPAN-MV330-60Hz", "Japan"),
    367: GridCode("JAPAN-MV380-60Hz", "Japan"),
    368: GridCode("LITHUANIA-EN50549-MV800", "Lithuania"),
    369: GridCode("FINLAND-EN50549-MV400", "Finland"),
    370: GridCode("FINLAND-EN50549-MV480", "Finland"),
    371: GridCode("FINLAND-EN50549-MV800", "Finland"),
    372: GridCode("EN50549-SE-MV400", "Sweden"),
    373: GridCode("EN50549-SE-MV480", "Sweden"),
    374: GridCode("EN50549-SE-MV800", "Sweden"),
    375: GridCode("CEPM", "Dominican Republic"),
    376: GridCode("CEPM-MV480", "Dominican Republic"),
    377: GridCode("CEPM-MV800", "Dominican Republic"),
    378: GridCode("SA-BESF-L", "South Africa"),
    379: GridCode("SA-BESF-L-MV480", "South Africa"),
    380: GridCode("SA-BESF-L-MV800", "South Africa"),
    381: GridCode("SA-BESF-H", "South Africa"),
    382: GridCode("SA-BESF-H-MV480", "South Africa"),
    383: GridCode("SA-BESF-H-MV800", "South Africa"),
    384: GridCode("BRAZIL-P140-LV220", "Brzail"),
    385: GridCode("NEW CALEDONIA-LV230", "New Caledonia"),
    386: GridCode("Israel-MV400", "Israel"),
    387: GridCode("ANRE-TYPEB", "Romania"),
    388: GridCode("ANRE-TYPEB-MV480", "Romania"),
    389: GridCode("AUSTRIANER_TYPEB_LV400", "Austria"),
    390: GridCode("AUSTRIANER_TYPEB_LV480", "Austria"),
    391: GridCode("AUSTRIANER_TYPEB_MV400", "Austria"),
    392: GridCode("AUSTRIANER_TYPEB_MV480", "Austria"),
    393: GridCode("IRAQ-MV800", "Iraq"),
    394: GridCode("MOROCCO-MV800", "Morocco"),
    395: GridCode("ALGERIA-MV800", "Algeria"),
    396: GridCode("CHINA-GBT19964-MV800", "China"),
    397: GridCode("CHINA-GBT29319-MV800", "China"),
    398: GridCode("SENEGAL", "Senegal"),
    399: GridCode("SENEGAL-MV480", "Senegal"),
    400: GridCode("SENEGAL-MV800", "Senegal"),
    401: GridCode("NC2022", "New Caledonia"),
}

STATE_CODES_1 = {
    0b0000_0000_0000_0001: "Standby",
    0b0000_0000_0000_0010: "Grid-Connected",
    0b0000_0000_0000_0100: "Grid-Connected normally",
    0b0000_0000_0000_1000: "Grid connection with derating due to power rationing",
    0b0000_0000_0001_0000: "Grid connection with derating due to internal causes of the solar inverter",
    0b0000_0000_0010_0000: "Normal stop",
    0b0000_0000_0100_0000: "Stop due to faults",
    0b0000_0000_1000_0000: "Stop due to power rationing",
    0b0000_0001_0000_0000: "Shutdown",
    0b0000_0010_0000_0000: "Spot check",
}

STATE_CODES_2 = {
    0b0000_0000_0000_0001: OnOffBit("Locked", "Unlocked"),
    0b0000_0000_0000_0010: OnOffBit("PV disconnected", "PV connected"),
    0b0000_0000_0000_0100: OnOffBit("No DSP data collection", "DSP data collection"),
}

STATE_CODES_3 = {
    0b0000_0000_0000_0000_0000_0000_0000_0001: OnOffBit("On-grid", "Off-grid"),
    0b0000_0000_0000_0000_0000_0000_0000_0010: OnOffBit(
        "Off-grid switch disabled",
        "Off-grid switch enabled",
    ),
}

ALARM_CODES_1 = {
    0b0000_0000_0000_0001: Alarm("High String Input Voltage", 2001, "Major"),
    0b0000_0000_0000_0010: Alarm("DC Arc Fault", 2002, "Major"),
    0b0000_0000_0000_0100: Alarm("String Reverse Connection", 2011, "Major"),
    0b0000_0000_0000_1000: Alarm("String Current Backfeed", 2012, "Warning"),
    0b0000_0000_0001_0000: Alarm("Abnormal String Power", 2013, "Warning"),
    0b0000_0000_0010_0000: Alarm("AFCI Self-Check Fail", 2021, "Major"),
    0b0000_0000_0100_0000: Alarm("Phase Wire Short-Circuited to PE", 2031, "Major"),
    0b0000_0000_1000_0000: Alarm("Grid Loss", 2032, "Major"),
    0b0000_0001_0000_0000: Alarm("Grid Undervoltage", 2033, "Major"),
    0b0000_0010_0000_0000: Alarm("Grid Overvoltage", 2034, "Major"),
    0b0000_0100_0000_0000: Alarm("Grid Volt. Imbalance", 2035, "Major"),
    0b0000_1000_0000_0000: Alarm("Grid Overfrequency", 2036, "Major"),
    0b0001_0000_0000_0000: Alarm("Grid Underfrequency", 2037, "Major"),
    0b0010_0000_0000_0000: Alarm("Unstable Grid Frequency", 2038, "Major"),
    0b0100_0000_0000_0000: Alarm("Output Overcurrent", 2039, "Major"),
    0b1000_0000_0000_0000: Alarm("Output DC Component Overhigh", 2040, "Major"),
}

ALARM_CODES_2 = {
    0b0000_0000_0000_0001: Alarm("Abnormal Residual Current", 2051, "Major"),
    0b0000_0000_0000_0010: Alarm("Abnormal Grounding", 2061, "Major"),
    0b0000_0000_0000_0100: Alarm("Low Insulation Resistance", 2062, "Major"),
    0b0000_0000_0000_1000: Alarm("Overtemperature", 2063, "Minor"),
    0b0000_0000_0001_0000: Alarm("Device Fault", 2064, "Major"),
    0b0000_0000_0010_0000: Alarm("Upgrade Failed or Version Mismatch", 2065, "Minor"),
    0b0000_0000_0100_0000: Alarm("License Expired", 2066, "Warning"),
    0b0000_0000_1000_0000: Alarm("Faulty Monitoring Unit", 61440, "Minor"),
    0b0000_0001_0000_0000: Alarm("Faulty Power Collector", 2067, "Major"),
    0b0000_0010_0000_0000: Alarm("Battery abnormal", 2068, "Minor"),
    0b0000_0100_0000_0000: Alarm("Active Islanding", 2070, "Major"),
    0b0000_1000_0000_0000: Alarm("Passive Islanding", 2071, "Major"),
    0b0001_0000_0000_0000: Alarm("Transient AC Overvoltage", 2072, "Major"),
    0b0010_0000_0000_0000: Alarm("Peripheral port short circuit", 2075, "Warning"),
    0b0100_0000_0000_0000: Alarm("Churn output overload", 2077, "Major"),
    0b1000_0000_0000_0000: Alarm("Abnormal PV module configuration", 2080, "Major"),
}

ALARM_CODES_3 = {
    0b0000_0000_0000_0001: Alarm("Optimizer fault", 2081, "Warning"),
    0b0000_0000_0000_0010: Alarm("Built-in PID operation abnormal", 2085, "Minor"),
    0b0000_0000_0000_0100: Alarm("High input string voltage to ground", 2014, "Major"),
    0b0000_0000_0000_1000: Alarm("External Fan Abnormal", 2086, "Major"),
    0b0000_0000_0001_0000: Alarm("Battery Reverse Connection", 2069, "Major"),
    0b0000_0000_0010_0000: Alarm("On-grid/Off-grid controller abnormal", 2082, "Major"),
    0b0000_0000_0100_0000: Alarm("PV String Loss", 2015, "Warning"),
    0b0000_0000_1000_0000: Alarm("Internal Fan Abnormal", 2087, "Major"),
    0b0000_0001_0000_0000: Alarm("DC Protection Unit Abnormal", 2088, "Major"),
    0b0000_0010_0000_0000: Alarm("EL Unit Abnormal", 2089, "Minor"),
    0b0000_0100_0000_0000: Alarm(
        "Active Adjustment Instruction Abnormal",
        2090,
        "Major",
    ),
    0b0000_1000_0000_0000: Alarm(
        "Reactive Adjustment Instruction Abnormal",
        2091,
        "Major",
    ),
    0b0001_0000_0000_0000: Alarm("CT Wiring Abnormal", 2092, "Major"),
    0b0010_0000_0000_0000: Alarm(
        "DC Arc Fault(ADMC Alarm to be clear manually)",
        2003,
        "Major",
    ),
    0b0100_0000_0000_0000: Alarm("DC Switch Abnormal", 2093, "Minor"),
    0b1000_0000_0000_0000: Alarm(
        "Allowable discharge capacity of the battery is low",
        2094,
        "Warning",
    ),
}


class StorageCapacityControlMode(IntEnum):
    """Storage Capacity Control Mode."""

    DISABLE = 0
    ACTIVE_CAPACITY_CONTROL = 1
    APPARENT_POWER_LIMIT = 2


class WlanWakeup(IntEnum):
    """WLAN Wakeup."""

    WAKEN_UP = 0
    DISABLED = 1


class RemoteChargeDischargeControlMode(IntEnum):
    """Remote Charge/Discharge Control Mode."""

    LOCAL_CONTROL = 0
    REMOTE_CONTROL_MAXIMUM_SELF_CONSUMPTION = 1
    REMOTE_CONTROL_FULLY_FED_TO_GRID = 2
    REMOTE_CONTROL_TOU = 3
    REMOTE_CONTROL_AI_CONTROL = 4


class EmmaExternalMeterRunningStatus(IntEnum):
    """Emma External Meter Running Status."""

    ONLINE = 0
    OFFLINE = 1


class EmmaEssControlMode(_IntEnumWithPrettyString):
    """Emma ESS Control Mode."""

    RESERVED_1 = 1
    MAXIMUM_SELF_CONSUMPTION = 2
    RESERVED_3 = 3
    FULLY_FED_TO_GRID = 4
    TIME_OF_USE = 5
    THIRD_PARTY_DISPATCH = 6


class EmmaLimitationMode(_IntEnumWithPrettyString):
    """Emma Limitation mode."""

    TOTAL_POWER = 0
    SINGLE_PHASE_POWER = 1


class EmmaPowerSupplyConfiguration(_IntEnumWithPrettyString):
    """EMMA Power Supply Configuration."""

    NONE = 0
    MAINS_ONLY = 1
    MAINS_AND_GENERATOR = 2
    GENERATOR_ONLY = 3


class EmmaConsiderMainsFaultyIf(IntEnum):
    """Consider mains to be faulty if."""

    OPEN = 0
    CLOSED = 1



"""Exceptions from the Huawei Solar library."""


class HuaweiSolarException(Exception):
    """Base class for Huawei Solar exceptions."""


class DecodeError(HuaweiSolarException):
    """Decoding failed."""


class EncodeError(HuaweiSolarException):
    """Encoding failed."""


class TimeOfUsePeriodsException(HuaweiSolarException):
    """Validation of TOU periods failed."""


class PeakPeriodsValidationError(Exception):
    """Validation of Peak Periods failed."""


class ConnectionException(HuaweiSolarException):
    """Exception connecting to device."""


class ReadException(HuaweiSolarException):
    """Exception reading register from device."""

    def __init__(self, *args, modbus_exception_code: int | None = None, **kwargs):
        """Create ReadException."""
        super().__init__(*args, **kwargs)
        self.modbus_exception_code = modbus_exception_code


class ConnectionInterruptedException(HuaweiSolarException):
    """Connection to the inverter was interrupted."""


class SlaveBusyException(HuaweiSolarException):
    """Non-fatal exception while trying to read from device."""


class SlaveFailureException(HuaweiSolarException):
    """Possibly fatal exception while trying to read from device."""


class WriteException(HuaweiSolarException):
    """Exception writing register to device."""

    def __init__(self, *args, modbus_exception_code: int | None = None, **kwargs):
        """Create WriteException."""
        super().__init__(*args, **kwargs)
        self.modbus_exception_code = modbus_exception_code


class PermissionDenied(HuaweiSolarException):
    """The inverter returned an error indicating that you don't have permission for this action."""


class InvalidCredentials(HuaweiSolarException):
    """Logging in on the inverter failed."""



"""Low-level Modbus logic."""

import asyncio
import hmac
import logging
import secrets
import struct
import sys
import time
import typing as t
from contextlib import asynccontextmanager
from dataclasses import dataclass
from hashlib import sha256
from typing import cast

import backoff
from pymodbus.constants import Endian
from pymodbus.exceptions import ConnectionException as ModbusConnectionException
from pymodbus.message.rtu import MessageRTU
from pymodbus.payload import BinaryPayloadBuilder, BinaryPayloadDecoder
from pymodbus.pdu import ExceptionResponse, ModbusExceptions, ModbusRequest
from pymodbus.register_write_message import (
    WriteMultipleRegistersResponse,
    WriteSingleRegisterResponse,
)

from .const import DEVICE_INFOS_START_OBJECT_ID, MAX_BATCHED_REGISTERS_COUNT
from .exceptions import (
    ConnectionException,
    ConnectionInterruptedException,
    HuaweiSolarException,
    PermissionDenied,
    ReadException,
    SlaveBusyException,
    SlaveFailureException,
    WriteException,
)
from .modbus import (
    AsyncHuaweiSolarModbusSerialClient,
    AsyncHuaweiSolarModbusTcpClient,
    CompleteUploadModbusRequest,
    CompleteUploadModbusResponse,
    PrivateHuaweiModbusRequest,
    PrivateHuaweiModbusResponse,
    ReadDeviceIdentifierRequest,
    ReadDeviceIdentifierResponse,
    StartUploadModbusRequest,
    StartUploadModbusResponse,
    UploadModbusRequest,
    UploadModbusResponse,
)
from .registers import REGISTERS, RegisterDefinition

LOGGER = logging.getLogger(__name__)


class Result(t.NamedTuple):
    """Modbus register value."""

    value: t.Any
    unit: str | None


DEFAULT_TCP_PORT = 502
DEFAULT_BAUDRATE = 9600

DEFAULT_SLAVE_ID = 0
DEFAULT_TIMEOUT = 10  # especially the SDongle can react quite slowly
DEFAULT_WAIT = 1
DEFAULT_COOLDOWN_TIME = 0.05
WAIT_FOR_CONNECTION_TIMEOUT = 5

HEARTBEAT_REGISTER = 49999

FILE_UPLOAD_MAX_RETRIES = 6
FILE_UPLOAD_RETRY_TIMEOUT = 10

PERMISSION_DENIED_EXCEPTION_CODE = 0x80

LOGIN_CHALLENGE_SUBCOMMAND = 0x11


def _compute_digest(password, seed):
    hashed_password = sha256(password).digest()

    return hmac.digest(key=hashed_password, msg=seed, digest=sha256)


@dataclass(frozen=True)
class DeviceInfo:
    """Device information."""

    model: str | None
    software_version: str | None
    interface_protocol_version: str | None
    esn: str | None
    device_id: int | None
    feature_version: str | None
    unknown_field: str | None
    product_type: str | None


@dataclass(frozen=True)
class DeviceIdentifier:
    """Device identifier information."""

    vendor: str
    product_code: str
    main_revision_version: str
    other_data: dict[int, bytes]


class AsyncHuaweiSolar:
    """Async interface to the Huawei solar inverter."""

    _reconnect_task: asyncio.Task | None = None
    __last_call_finished_at: float | None = None

    def __init__(
        self,
        client: AsyncHuaweiSolarModbusSerialClient | AsyncHuaweiSolarModbusTcpClient,
        slave_id: int = DEFAULT_SLAVE_ID,
        timeout: int = DEFAULT_TIMEOUT,
        cooldown_time: float = DEFAULT_COOLDOWN_TIME,
    ):
        """DO NOT USE THIS CONSTRUCTOR DIRECTLY. Use AsyncHuaweiSolar.create() instead."""
        self._client = client
        self._timeout = timeout
        self._cooldown_time = cooldown_time
        self.slave_id = slave_id

        # use this lock to prevent concurrent requests, as the
        # Huawei inverters can't cope with those
        self.__communication_lock = asyncio.Lock()

    @asynccontextmanager
    async def _communication_lock(self):
        async with self.__communication_lock:
            if not self._client.connected_event.is_set():
                LOGGER.info("Waiting for connection ...")

            try:
                await asyncio.wait_for(
                    self._client.connected_event.wait(),
                    WAIT_FOR_CONNECTION_TIMEOUT,
                )
            except asyncio.TimeoutError:
                LOGGER.exception(
                    "Timeout while waiting for connection. Reconnecting...",
                )
                self._reconnect_task = asyncio.create_task(self._reconnect())
                raise

            if self.__last_call_finished_at:
                cooldown_time_needed = (self.__last_call_finished_at + self._cooldown_time) - time.time()
                if cooldown_time_needed > 0:
                    LOGGER.debug(
                        "Sleeping for %f seconds before making next call.",
                        cooldown_time_needed,
                    )
                    await asyncio.sleep(cooldown_time_needed)

            try:
                yield
            finally:
                self.__last_call_finished_at = time.time()

    @classmethod
    async def create(  # noqa: PLR0913
        cls,
        host,
        port: int = DEFAULT_TCP_PORT,
        slave: int = DEFAULT_SLAVE_ID,
        timeout: int = DEFAULT_TIMEOUT,
        cooldown_time: float = DEFAULT_COOLDOWN_TIME,
    ):
        """Create an AsyncHuaweiSolar instance."""
        client = None
        try:
            client = AsyncHuaweiSolarModbusTcpClient(host, port, timeout)
            await client.connect()

            huawei_solar = cls(client, slave, timeout, cooldown_time)
        except Exception as err:
            # if an error occurs, we need to make sure that the Modbus-client is stopped,
            # otherwise it can stay active and cause even more problems ...
            LOGGER.exception("Aborting client creation due to error.")

            if client:
                client.close()
            raise ConnectionException from err
        else:
            return huawei_solar

    @classmethod
    async def create_rtu(  # noqa: PLR0913
        cls,
        port,
        baudrate: int = DEFAULT_BAUDRATE,
        slave: int = DEFAULT_SLAVE_ID,
        timeout: int = DEFAULT_TIMEOUT,
        cooldown_time: float = DEFAULT_COOLDOWN_TIME,
        **serial_kwargs,
    ):
        """Create a serial client."""
        client = None
        try:
            client = AsyncHuaweiSolarModbusSerialClient(
                port,
                baudrate,
                timeout,
                **serial_kwargs,
            )
            await client.connect()

            # wait a little bit to prevent a timeout on the first request
            await asyncio.sleep(1)

            huawei_solar = cls(client, slave, timeout, cooldown_time)
        except Exception as err:
            # if an error occurs, we need to make sure that the Modbus-client is stopped,
            # otherwise it can stay active and cause even more problems ...
            LOGGER.exception("Aborting client creation due to error.")
            raise ConnectionException from err
        else:
            return huawei_solar

    async def stop(self):
        """Stop the modbus client."""
        if self._reconnect_task:
            self._reconnect_task.cancel()

        self._client.close()

    async def _reconnect(self):
        """Reconnect to the inverter."""
        self._client.close()
        await self._client.connect()

    async def _decode_response(
        self,
        reg: RegisterDefinition,
        decoder: BinaryPayloadDecoder,
    ):
        """Decode a modbus register and puts it into a Result object."""
        result = reg.decode(decoder)

        if not hasattr(reg, "unit") or callable(reg.unit) or isinstance(reg.unit, dict):
            return Result(result, None)
        return Result(result, reg.unit)

    async def get(self, name, slave=None):
        """Get named register from device."""
        return (await self.get_multiple([name], slave))[0]

    async def get_multiple(self, names: list[str], slave=None):
        """Read multiple registers at the same time.

        This is only possible if the registers are consecutively available in the
        inverters' memory.
        """
        if len(names) == 0:
            raise ValueError("Expected at least one register name")

        registers = list(map(REGISTERS.get, names))

        if None in registers:
            raise ValueError("Did not recognize all register names")
        registers = t.cast(list[RegisterDefinition], registers)

        for register, register_name in zip(registers, names, strict=False):
            if not register.readable:
                raise ValueError(f"Trying to read unreadable register {register_name}")

        for idx in range(1, len(names)):
            if registers[idx - 1].register + registers[idx - 1].length > registers[idx].register:
                raise ValueError(
                    f"Requested registers must be in monotonically increasing order, "
                    f"but {registers[idx-1].register} + {registers[idx-1].length} > {registers[idx].register}!",
                )

            register_distance = registers[idx - 1].register + registers[idx - 1].length - registers[idx].register

            if register_distance > MAX_BATCHED_REGISTERS_COUNT:
                raise ValueError(
                    "Gap between requested registers is too large. Split it in two requests",
                )

        total_length = registers[-1].register + registers[-1].length - registers[0].register

        response = await self._read_registers(
            registers[0].register,
            total_length,
            slave,
        )

        decoder = BinaryPayloadDecoder.fromRegisters(
            response.registers,
            byteorder=Endian.BIG,
            wordorder=Endian.BIG,
        )

        result = [await self._decode_response(registers[0], decoder)]
        for idx in range(1, len(registers)):
            skip_registers = registers[idx].register - (registers[idx - 1].register + registers[idx - 1].length)
            decoder.skip_bytes(
                skip_registers * 2,
            )  # registers are 16-bit, so we need to multiply by two
            result.append(await self._decode_response(registers[idx], decoder))

        return result

    async def _read_registers(  # noqa: C901
        self,
        register: int,
        length: int,
        slave: int | None,
    ):
        """Async read register from device.

        The device needs a bit of time between the connection and the first request
        and between requests if there is a long time between them, else it will fail.

        This is solved by sleeping between the first connection and a request,
        and up to 5 retries between following requests.

        It seems to only support connections from one device at the same time.
        """

        def on_backoff(details):
            LOGGER.debug(
                "Received %s: backing off reading for %0.1f seconds after %d tries",
                sys.exc_info()[0],
                details["wait"],
                details["tries"],
            )

        def on_backoff_with_reconnect(details):
            if details["tries"] % 3 == 0:
                self._reconnect_task = asyncio.create_task(self._reconnect())
                LOGGER.debug(
                    "Received %s: reconnecting and backing off reading for %0.1f seconds after %d tries",
                    sys.exc_info()[0],
                    details["wait"],
                    details["tries"],
                )
            else:
                LOGGER.debug(
                    "Received %s: backing off reading for %0.1f seconds after %d tries",
                    sys.exc_info()[0],
                    details["wait"],
                    details["tries"],
                )

        def backoff_giveup(details):
            raise ReadException(
                f"Failed to read register {register} after {details['tries']} tries",
            )

        @backoff.on_exception(
            backoff.expo,
            (asyncio.TimeoutError, ConnectionInterruptedException),
            max_tries=6,
            jitter=None,
            on_backoff=on_backoff_with_reconnect,
            on_giveup=backoff_giveup,
        )
        @backoff.on_exception(
            backoff.expo,
            (SlaveBusyException, SlaveFailureException),
            max_tries=6,
            jitter=None,
            on_backoff=on_backoff,
            on_giveup=backoff_giveup,
        )
        async def _do_read():
            if not self._client.connected:
                message = "Modbus client is not connected to the inverter."
                LOGGER.exception(message)
                raise ConnectionInterruptedException(message)
            try:
                response = await self._client.read_holding_registers(
                    register,
                    length,
                    slave=slave or self.slave_id,
                )

                # trigger a backoff if we get a SlaveBusy-exception
                if isinstance(response, ExceptionResponse):
                    if response.exception_code == ModbusExceptions.SlaveBusy:
                        LOGGER.debug(
                            "Got a SlaveBusy Modbus Exception while reading %d (length %d) from slave %d",
                            register,
                            length,
                            slave or self.slave_id,
                        )
                        raise SlaveBusyException

                    if response.exception_code == ModbusExceptions.SlaveFailure:
                        LOGGER.debug(
                            "Got a SlaveFailure Modbus Exception while reading %d (length %d) from slave %d",
                            register,
                            length,
                            slave or self.slave_id,
                        )
                        raise SlaveFailureException

                    # Not a SlaveBusy or SlaveFailure exception
                    raise ReadException(
                        f"Got error while reading from register {register} with length {length}: {response}",
                        modbus_exception_code=response.exception_code,
                    )

                if len(response.registers) != length:
                    raise SlaveBusyException(
                        f"Mismatch between number of requested registers ({length}) "
                        f"and number of received registers ({len(response.registers)})",
                    )

            except ModbusConnectionException as err:
                message = "Could not read register value, has another device interrupted the connection?"
                LOGGER.exception(message)
                raise ConnectionInterruptedException(message) from err
            else:
                return response

        async with self._communication_lock():
            LOGGER.debug(
                "Reading register %d with length %d from slave %s",
                register,
                length,
                slave or self.slave_id,
            )
            return await _do_read()

    async def _read_device_identifier_objects(self, read_dev_id_code: int, object_id: int):
        """Read all the objects of a certain ReadDevId code."""
        next_object_id: int | None = object_id

        objects = {}

        while next_object_id is not None:
            response = await self._client.execute(
                ReadDeviceIdentifierRequest(
                    read_dev_id_code=read_dev_id_code,
                    object_id=next_object_id,
                ),
            )

            if isinstance(response, ExceptionResponse):
                if response.exception_code == PERMISSION_DENIED_EXCEPTION_CODE:
                    raise PermissionDenied("Permission denied")
                if response.exception_code == ModbusExceptions.SlaveBusy:
                    raise SlaveBusyException
                if response.exception_code == ModbusExceptions.SlaveFailure:
                    raise SlaveFailureException
                raise ReadException(
                    f"Exception occurred while trying to read device infos "
                    f"{hex(response.exception_code) if response.exception_code else 'no exception code'}",
                    modbus_exception_code=response.exception_code,
                )

            assert isinstance(response, ReadDeviceIdentifierResponse)
            objects.update(response.objects)
            next_object_id = response.next_object_id if response.more else None

        return objects

    async def get_device_identifiers(self):
        """Read the device identifiers from the inverter."""
        objects = await self._read_device_identifier_objects(0x01, 0x00)

        return DeviceIdentifier(
            vendor=objects.pop(0x00).decode("ascii"),
            product_code=objects.pop(0x01).decode("ascii"),
            main_revision_version=objects.pop(0x02).decode("ascii"),
            other_data=objects,
        )

    async def get_device_infos(self):
        """Read the device infos from the inverter."""
        objects = await self._read_device_identifier_objects(0x03, DEVICE_INFOS_START_OBJECT_ID)

        def _parse_device_entry(device_info_str: str) -> DeviceInfo:
            raw_device_info: dict[int, str] = {}
            for entry in device_info_str.split(";"):
                key, value = entry.split("=")
                raw_device_info[int(key)] = value

            return DeviceInfo(
                model=raw_device_info.get(1),
                software_version=raw_device_info.get(2),
                interface_protocol_version=raw_device_info.get(3),
                esn=raw_device_info.get(4),
                device_id=int(raw_device_info[5]) if 5 in raw_device_info else None,  # noqa: PLR2004
                feature_version=raw_device_info.get(6),
                unknown_field=raw_device_info.get(7),
                product_type=raw_device_info.get(8),
            )

        if DEVICE_INFOS_START_OBJECT_ID in objects:
            (number_of_devices,) = struct.unpack(">B", objects.pop(DEVICE_INFOS_START_OBJECT_ID))
        else:
            LOGGER.warning("No 0x87 entry with number of devices found in objects. Ignoring")
            number_of_devices = -1

        device_infos = [
            _parse_device_entry(device_info_bytes.decode("ascii")) for device_info_bytes in objects.values()
        ]

        if number_of_devices >= 0 and len(device_infos) != number_of_devices:
            LOGGER.warning(
                "Number of device infos does not match the number of devices: %d != %d",
                len(device_infos),
                number_of_devices,
            )

        return device_infos

    async def get_file(
        self,
        file_type,
        customized_data=None,
        slave: int | None = None,
    ) -> bytes:
        """Read a 'file' via Modbus.

        As defined by the 'Uploading Files' process described in 6.3.7.1 of
        the Solar Inverter Modbus Interface Definitions PDF.
        """

        def backoff_giveup(details):
            raise ReadException(
                f"Failed to read file {file_type} after {details['tries']} tries",
            )

        @backoff.on_exception(
            backoff.constant,
            (asyncio.TimeoutError, SlaveBusyException, SlaveFailureException),
            interval=FILE_UPLOAD_RETRY_TIMEOUT,
            max_tries=FILE_UPLOAD_MAX_RETRIES,
            jitter=None,
            on_backoff=lambda details: LOGGER.debug(
                "Received %s: backing off file upload for %0.1f seconds after %d tries",
                sys.exc_info()[0],
                details.get("wait"),
                details["tries"],
            ),
            on_giveup=backoff_giveup,
        )
        async def _perform_request(request: ModbusRequest, response_type):
            response = cast(
                PrivateHuaweiModbusResponse,
                await self._client.execute(request),
            )

            if isinstance(response, ExceptionResponse):
                if response.exception_code == PERMISSION_DENIED_EXCEPTION_CODE:
                    raise PermissionDenied("Permission denied")
                if response.exception_code == ModbusExceptions.SlaveBusy:
                    raise SlaveBusyException
                if response.exception_code == ModbusExceptions.SlaveFailure:
                    raise SlaveFailureException
                raise ReadException(
                    f"Exception occurred while trying to read file {hex(file_type)}: "
                    f"{hex(response.exception_code) if response.exception_code else 'no exception code'}",
                    modbus_exception_code=response.exception_code,
                )

            return response_type(response.content)

        async def _do_read_file():
            # Start the upload
            start_upload_response = await _perform_request(
                StartUploadModbusRequest(
                    file_type,
                    customized_data,
                    slave=slave or self.slave_id,
                ),
                StartUploadModbusResponse,
            )

            data_frame_length = start_upload_response.data_frame_length
            file_length = start_upload_response.file_length

            # Request the data in 'frames'

            file_data = b""
            next_frame_no = 0

            while (next_frame_no * data_frame_length) < file_length:
                data_upload_response = await _perform_request(
                    UploadModbusRequest(
                        file_type,
                        next_frame_no,
                        slave=slave or self.slave_id,
                    ),
                    UploadModbusResponse,
                )

                file_data += data_upload_response.frame_data
                next_frame_no += 1

            # Complete the upload and check the CRC
            complete_upload_response = await _perform_request(
                CompleteUploadModbusRequest(file_type, slave=slave or self.slave_id),
                CompleteUploadModbusResponse,
            )

            file_crc = complete_upload_response.file_crc
            # swap upper and lower two bytes to match how computeCRC works
            swapped_crc = ((file_crc << 8) & 0xFF00) | ((file_crc >> 8) & 0x00FF)

            if not MessageRTU.check_CRC(file_data, swapped_crc):
                raise ReadException(
                    f"Computed CRC {MessageRTU.compute_CRC(file_data):x} for file {file_type} "
                    f"does not match expected value {swapped_crc}",
                )

            return file_data

        async with self._communication_lock():
            LOGGER.debug(
                "Reading file %#x from slave %d",
                file_type,
                slave or self.slave_id,
            )
            return await _do_read_file()

    async def set(self, name, value, slave=None):
        """Set named register on device."""
        try:
            reg = REGISTERS[name]
        except KeyError as err:
            raise ValueError("Invalid Register Name") from err

        if not reg.writeable:
            raise WriteException("Register is not writable")

        builder = BinaryPayloadBuilder(byteorder=Endian.BIG, wordorder=Endian.BIG)
        reg.encode(value, builder)
        value = builder.to_registers()

        if len(value) != reg.length:
            raise WriteException("Wrong number of registers to write")

        def backoff_giveup(details):
            raise ReadException(
                f"Failed to write to register after {details['tries']} tries",
            )

        @backoff.on_exception(
            backoff.expo,
            (asyncio.TimeoutError, SlaveBusyException, ConnectionInterruptedException),
            max_tries=3,
            jitter=None,
            on_backoff=lambda details: LOGGER.debug(
                "Received %s: backing off writing for %0.1f seconds after %d tries",
                sys.exc_info()[0],
                details.get("wait"),
                details["tries"],
            ),
            on_giveup=backoff_giveup,
        )
        async def _do_set():
            return await self._write_registers(
                reg.register,
                builder.to_registers(),
                slave,
            )

        async with self._communication_lock():
            LOGGER.debug(
                "Writing to register %s value %s on slave %s",
                name,
                value,
                slave or self.slave_id,
            )
            return await _do_set()

    async def _write_registers(
        self,
        register: int,
        value: list[int],
        slave=None,
    ) -> bool:
        """Async write register to device."""
        if not self._client.connected:
            message = "Modbus client is not connected to the inverter."
            LOGGER.exception(message)
            raise ConnectionInterruptedException(message)
        try:
            LOGGER.debug(
                "Writing to %d: %s on slave %d",
                register,
                value,
                slave or self.slave_id,
            )

            single_register = len(value) == 1
            if single_register:
                response = await self._client.write_register(
                    register,
                    value[0],
                    slave=slave or self.slave_id,
                )

            else:
                response = await self._client.write_registers(
                    register,
                    value,
                    slave=slave or self.slave_id,
                )

            if isinstance(response, ExceptionResponse):
                if response.exception_code == PERMISSION_DENIED_EXCEPTION_CODE:
                    raise PermissionDenied("Permission denied")
                if response.exception_code == ModbusExceptions.IllegalAddress:
                    # cfr. https://github.com/wlcrs/huawei_solar/issues/587
                    raise PermissionDenied(
                        f"Failed to write value {value} to register {register} due to IllegalAddress. "
                        "Assuming permission problem.",
                    )
                raise WriteException(
                    f"Failed to write value {value} to register {register}: "
                    f"{ModbusExceptions.decode(response.exception_code)}",
                    modbus_exception_code=response.exception_code,
                )

            if single_register:
                assert isinstance(response, WriteSingleRegisterResponse)
                return response.address == register and response.value == value[0]
            assert isinstance(response, WriteMultipleRegistersResponse)
            return response.address == register and response.count == len(value)
        except ModbusConnectionException as err:
            LOGGER.exception("Failed to connect to device, is the host correct?")
            raise ConnectionInterruptedException(err) from err

    async def login(self, username: str, password: str, slave: int | None = None) -> bool:
        """Login into the inverter."""

        def backoff_giveup(details):
            raise ReadException(f"Failed to login after {details['tries']} tries")

        @backoff.on_exception(
            backoff.expo,
            (
                asyncio.TimeoutError,
                SlaveBusyException,
                SlaveFailureException,
                SlaveFailureException,
            ),
            max_tries=4,
            jitter=None,
            on_backoff=lambda details: LOGGER.debug(
                "Received %s: backing off login for %0.1f seconds after %d tries",
                sys.exc_info()[0],
                details.get("wait"),
                details["tries"],
            ),
            on_giveup=backoff_giveup,
        )
        async def _do_login() -> bool:
            # Get challenge
            challenge_request = PrivateHuaweiModbusRequest(
                36,
                bytes([1, 0]),
                slave=slave or self.slave_id,
            )

            challenge_response = cast(
                PrivateHuaweiModbusResponse,
                await self._client.execute(challenge_request),
            )

            assert challenge_response.content[0] == LOGIN_CHALLENGE_SUBCOMMAND
            inverter_challenge = challenge_response.content[1:17]

            client_challenge = secrets.token_bytes(16)

            encoded_username = username.encode("utf-8")
            hashed_password = _compute_digest(
                password.encode("utf-8"),
                inverter_challenge,
            )

            login_bytes = bytes(
                [
                    len(client_challenge) + 1 + len(encoded_username) + 1 + len(hashed_password),
                    *client_challenge,
                    len(encoded_username),
                    *encoded_username,
                    len(hashed_password),
                    *hashed_password,
                ],
            )
            await asyncio.sleep(0.05)
            login_request = PrivateHuaweiModbusRequest(
                37,
                login_bytes,
                slave=slave or self.slave_id,
            )
            login_response = cast(
                PrivateHuaweiModbusResponse,
                await self._client.execute(login_request),
            )

            if login_response.content[1] == 0:
                # check if inverter returned the right hash of the password as well
                inverter_mac_response_lengths = login_response.content[2]

                inverter_mac_response = login_response.content[3 : 3 + inverter_mac_response_lengths]

                if _compute_digest(password.encode("utf-8"), client_challenge) != inverter_mac_response:
                    LOGGER.error(
                        "Inverter response contains an invalid challenge answer. This could indicate a MitM-attack!",
                    )

                return True
            return False

        async with self._communication_lock():
            LOGGER.debug("Logging in")
            return await _do_login()

    async def heartbeat(self, slave_id):
        """Perform the heartbeat command. Only useful when maintaining a session."""
        if not self._client.connected:
            return False
        try:
            # 49999 is the magic register used to keep the connection alive
            response = await self._client.write_register(
                HEARTBEAT_REGISTER,
                0x1,
                slave=slave_id or self.slave_id,
            )
            if isinstance(response, ExceptionResponse):
                LOGGER.warning(
                    "Received an error after sending the heartbeat command: %s",
                    ModbusExceptions.decode(response.exception_code),
                )
                return False
            LOGGER.debug("Heartbeat succeeded")
        except HuaweiSolarException:
            LOGGER.exception("Exception during heartbeat")
            return False
        else:
            return True



import huawei_solar.register_names as rn
from huawei_solar.registers import REGISTERS, PeakSettingPeriod
from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadBuilder, BinaryPayloadDecoder


def test_capacity_control_register():
    value = [
        PeakSettingPeriod(0, 1439, 2551, (True, True, True, True, True, True, False)),
        PeakSettingPeriod(
            0,
            200,
            2550,
            (False, False, False, False, False, False, True),
        ),
        PeakSettingPeriod(
            200,
            1439,
            2449,
            (False, False, False, False, False, False, True),
        ),
    ]

    pspr = REGISTERS[rn.STORAGE_CAPACITY_CONTROL_PERIODS]

    builder = BinaryPayloadBuilder(byteorder=Endian.BIG, wordorder=Endian.BIG)
    pspr.encode(value, builder)

    payload = builder.to_registers()

    decoder = BinaryPayloadDecoder.fromRegisters(
        payload,
        byteorder=Endian.BIG,
        wordorder=Endian.BIG,
    )

    decoded_result = pspr.decode(decoder)

    assert decoded_result == value
